import json

from dataclasses import asdict, dataclass, field
from enum import Enum
from typing import TypeAlias, Mapping, Any, Optional

from metasdk.tools.validators import check_postgres_datetime_without_tz


JSONb: TypeAlias = Mapping[str, Any]


class IssueStatus(str, Enum):
    WAITING = "1"
    IN_PROGRESS = "2"
    DONE = "3"
    CLARIFICATION = "4"
    PENDING = "5"
    DEPRECATED_PLANNED = "6"
    CANCELED = "7"
    APPROVED = "8"
    APPROVAL_DENIED = "9"


@dataclass
class BaseIssueItem:
    """
    Базовый Датакласс для хранения данных о тикете
    """
    name: str
    issue_type_id: str
    reporter_user_id: int
    entity_id: str
    object_id: str
    application_id: str
    description: str
    issue_priority_id: str
    issue_status_id: str
    assignee_user_id: int
    form_data: JSONb
    watcher_users: JSONb
    deadline_time: str
    estimated_time: float
    tags: list[str]

    def to_dict(self):
        return asdict(self)


@dataclass
class IssueCreateItem(BaseIssueItem):
    """
    Датакласс для хранения данных о создаваемом тикете
    deadline_time должно быть в формате "YYYY-MM-DDTHH:MM:SS"
    estimated_time передается в ЧАСАХ с дальнейшей конвертацией в минуты
    """
    description: str = ""
    issue_priority_id: str = "1"
    issue_status_id: str = "1"
    assignee_user_id: Optional[int] = None
    form_data: JSONb = field(default_factory=dict)
    watcher_users: JSONb = field(default_factory=dict)
    deadline_time: Optional[str] = None
    estimated_time: Optional[float] = None
    tags: list[str] = field(default_factory=list)

    @property
    def deadline_time_datetime(self) -> str | None:
        if self.deadline_time is not None:
            check_postgres_datetime_without_tz(self.deadline_time)
        return self.deadline_time

    @property
    def estimated_time_interval(self) -> str | None:
        if self.estimated_time is None:
            return self.estimated_time
        return f"{self.estimated_time*60} MINUTES"

    def __post_init__(self):
        for issue_field in self.__dataclass_fields__:
            if isinstance(getattr(self, issue_field), dict):
                setattr(self, issue_field, json.dumps(getattr(self, issue_field)))
            if isinstance(getattr(self, issue_field), list):
                initial = str(getattr(self, issue_field))
                final = initial.replace('[', '{').replace(']', '}').replace('"', "").replace("'", "")
                setattr(self, issue_field, final)


@dataclass
class IssueGetItem(BaseIssueItem):
    """
    Датакласс для хранения данных о полученном тикете
    """
    id: int  # noqa: A003
    last_user_id: int
    creation_time: str
    modification_time: str
    status_change_time: str
    ref_key: str
    result: JSONb


@dataclass
class IssueUpdateItem:
    name: Optional[str] = None
    description: Optional[str] = None
    deadline_time: Optional[str] = None
    issue_priority_id: Optional[str] = None
    tags: list[str] = field(default_factory=list)

    @property
    def deadline_time_datetime(self) -> str | None:
        if self.deadline_time is not None:
            check_postgres_datetime_without_tz(self.deadline_time)
        return self.deadline_time

    def __post_init__(self):
        for issue_field in self.__dataclass_fields__:
            if isinstance(getattr(self, issue_field), list):
                initial = str(getattr(self, issue_field))
                final = initial.replace('[', '{').replace(']', '}').replace('"', "").replace("'", "")
                setattr(self, issue_field, final)


from typing import Optional, Mapping, Any

from metasdk.schemas import IssueCreateItem, IssueGetItem, IssueUpdateItem, IssueStatus
from metasdk.exceptions import BadParametersError

from metasdk.tools.validators import check_postgres_date, check_postgres_period
from metasdk.tools.wrappers import common_auth
from metasdk.tools.parsers import get_bind_params


class IssueService:

    def __init__(self, app, token: Optional[str] = ""):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__options: Mapping[str, Any] = {}
        self.__data_get_cache: Mapping[str, Any] = {}
        self.__metadb = app.db("meta")
        self.token = token
        self.auth_service = app.AuthService
        self.required_scopes: list[str] = []
        self.auth_user_info: Optional[Mapping[str, Any]] = None

    def change_issue_status(self, issue_id: int, status_id: str) -> None:
        """
        Смета статуса тикета
        :param issue_id: Идентификатор тикета.
        :param status_id: Идентификатор статуса тикета.
        :return: None.
        """
        self.__metadb.update(
            """
            UPDATE meta.issue
            SET 
                issue_status_id = :status_id,
                assignee_user_id = valera_user_id(),
                last_user_id = valera_user_id()
            WHERE id = :issue_id
        """,
            {"issue_id": issue_id, "status_id": status_id},
        )

    def pending_issue(self, issue_id: int) -> None: 
        """
        Перевести тикет в статус "В ожидании"
        :param issue_id: int
        """
        self.change_issue_status(issue_id, "1")

    def in_progress_issue(self, issue_id: int) -> None:
        """
        Взять в работу
        :param issue_id: int
        """
        self.change_issue_status(issue_id, "2")

    def reject_issue(self, issue_id: int) -> None:
        """
        Отклонить задачу
        :param issue_id: int
        """
        self.change_issue_status(issue_id, "7")

    def clarification_issue(self, issue_id: int) -> None:
        """
        Уточнение
        :param issue_id: int
        """
        self.change_issue_status(issue_id, "4")

    def done_issue(self, issue_id: int) -> None:
        """
        Успешное выполенение
        :param issue_id: int
        """
        self.change_issue_status(issue_id, "3")

    def add_issue_msg(self, issue_id: int, msg: str) -> None:
        self.__metadb.update(
            """
            INSERT INTO meta.issue_msg (msg, issue_id, user_id, last_user_id)
            VALUES (:msg, :issue_id, valera_user_id(), valera_user_id())
        """,
            {"issue_id": issue_id, "msg": msg},
        )

    @common_auth
    def add_issue_msg_secure(self, issue_id: int, msg: str):
        if not msg.strip():
            raise BadParametersError("Message should not be empty")
        issue_msg = self.__metadb.query(
            """
            INSERT INTO meta.issue_msg (msg, issue_id, user_id, last_user_id)
            VALUES (:msg, :issue_id, :user_id, :user_id)
            RETURNING id, msg
        """,
            {
                "issue_id": issue_id,
                "msg": msg,
                "user_id": self.auth_user_info["auth_user_id"] if self.auth_user_info else None,
            },
        )
        issue_msg_id = issue_msg.get("rows", [])[0].get("id")
        issue_msg_msg = issue_msg.get("rows", [])[0].get("msg")
        return {"message_id": issue_msg_id, "message": issue_msg_msg}

    @common_auth
    def change_issue_status_secure(self, issue_id: int, status_id: str):
        if status_id not in IssueStatus:
            raise BadParametersError(
                f"Status id should be one of {IssueStatus._member_map_}"
            )
        issue = self.__metadb.query(
            """
              UPDATE meta.issue SET 
                issue_status_id=:status_id,
                assignee_user_id=:user_id,
                last_user_id=:user_id
              WHERE id = :issue_id::bigint
              RETURNING id, issue_status_id
        """,
            {
                "issue_id": issue_id,
                "status_id": status_id,
                "user_id": self.auth_user_info["auth_user_id"] if self.auth_user_info else None,
            },
        )
        issue_id = issue.get("rows", [])[0].get("id")
        issue_status_id = issue.get("rows", [])[0].get("issue_status_id")
        return {"issue_id": issue_id, "issue_status_id": issue_status_id}

    @common_auth
    def create_issue(self, issue: IssueCreateItem) -> int:
        """
        Создание тикета
        :param issue: IssueCreateItem объект с данными тикета
        :return: ticket_id int айди созданного тикета
        """
        query = """
            INSERT INTO meta.issue (
                application_id,
                name,
                description,
                form_data,
                issue_type_id,
                issue_priority_id,
                reporter_user_id,
                last_user_id,
                entity_id,
                object_id,
                deadline_time,
                estimated_time,
                tags,
                watcher_users,
                assignee_user_id)
            SELECT 
                :application_id::text,
                :issue_name::text,
                :description::text,
                :form_data::jsonb,
                :issue_type_id::text,
                :issue_priority_id::text,
                :reporter_user_id::bigint,
                :last_user_id::bigint,
                :entity_id::text,
                :object_id::text,
                :deadline_time::timestamp,
                :estimated_time::interval,
                :tags::text[],
                :watcher_users::jsonb,
                :assignee_user_id::bigint
            RETURNING id
        """
        ticket_id = self.__metadb.query(
            query,
            {
                "application_id": issue.application_id,
                "issue_name": issue.name,
                "description": issue.description,
                "form_data": issue.form_data,
                "issue_type_id": issue.issue_type_id,
                "issue_priority_id": issue.issue_priority_id,
                "reporter_user_id": issue.reporter_user_id,
                "last_user_id": self.auth_user_info["auth_user_id"] if self.auth_user_info else None,
                "entity_id": issue.entity_id,
                "object_id": issue.object_id,
                "deadline_time": issue.deadline_time_datetime,
                "estimated_time": issue.estimated_time_interval,
                "tags": issue.tags,
                "watcher_users": issue.watcher_users,
                "assignee_user_id": issue.assignee_user_id,
            },
        )
        return ticket_id.get("rows", [])[0].get("id")

    @common_auth
    def get_issue(self, issue_id: int) -> IssueGetItem | None:
        """
        Возвращает тикет по айди
        :param issue_id: int
        :return: IssueGetItem
        """
        try:
            issue_id = int(issue_id)
        except ValueError:
            raise BadParametersError("Issue id should be integer")
        query = """
            SELECT 
                id,
                name,
                description,
                form_data,
                creation_time,
                modification_time,
                last_user_id,
                issue_type_id,
                reporter_user_id,
                assignee_user_id,
                issue_priority_id,
                issue_status_id,
                entity_id,
                object_id,
                application_id,
                watcher_users,
                deadline_time,
                status_change_time,
                ref_key,
                estimated_time,
                tags,
                result
            FROM meta.issue WHERE id = :id
        """
        issue = self.__metadb.one(query, {"id": issue_id})
        return IssueGetItem(**issue) if issue else None

    @common_auth
    def get_issues(
        self,
        creation_time: Optional[str] = None,
        modification_time_start: Optional[str] = None,
        modification_time_end: Optional[str] = None,
        issue_type_id: Optional[str] = None,
        issue_status_id: Optional[str] = None,
        assignee_user_id: Optional[int] = None,
        reporter_user_id: Optional[int] = None,
        watcher_users: Optional[list[int]] = None,
        form_data: Optional[dict[str, Any]] = None,
        result: Optional[dict[str, Any]] = None
    ) -> list[IssueGetItem]:
        """
        Возвращает список тикетов. Выдача ограничена 1000 записями
        :param creation_time: str согласно выдаче метода DbQueryService.all соответсвует формату '%Y-%m-%dT%H:%M:%S.%f%z',
        но может использоваться формат '%Y-%m-%d', так как в запросе приводится к date
        :param modification_time: str согласно выдаче метода DbQueryService.all соответсвует формату '%Y-%m-%dT%H:%M:%S.%f%z',
        но может использоваться формат '%Y-%m-%d', так как в запросе приводится к date
        :param issue_type_id: str
        :param issue_status_id: str
        :param assignee_user_id: int
        :param reporter_user_id: int
        :param watcher_users: list[int]
        :return: list[IssueGetItem]
        """
        if watcher_users is None:
            watcher_users = []
        if creation_time is not None:
            check_postgres_date(creation_time)
        if modification_time_end and modification_time_start:
            check_postgres_period(modification_time_start, modification_time_end)
            modification_time_query = """
                AND modification_time::date BETWEEN :modification_time_start::date AND :modification_time_end::date
            """
        elif modification_time_start and not modification_time_end:
            check_postgres_date(modification_time_start)
            modification_time_query = """
                AND modification_time::date >= :modification_time_start::date
            """
        elif modification_time_end and not modification_time_start:
            check_postgres_date(modification_time_end)
            modification_time_query = """
                AND modification_time::date <= :modification_time_end::date
            """
        else:
            modification_time_query = ""
        query = """
            SELECT 
                id,
                name,
                description,
                form_data,
                creation_time,
                modification_time,
                last_user_id,
                issue_type_id,
                reporter_user_id,
                assignee_user_id,
                issue_priority_id,
                issue_status_id,
                entity_id,
                object_id,
                application_id,
                watcher_users,
                deadline_time,
                status_change_time,
                ref_key,
                estimated_time,
                tags,
                result
            FROM meta.issue
            WHERE 2>1
                AND creation_time::date = COALESCE(:creation_time::date, creation_time::date)
                AND issue_type_id::text = COALESCE(:issue_type_id::text, issue_type_id::text)
                AND issue_status_id::text = COALESCE(:issue_status_id::text, issue_status_id::text)
                AND assignee_user_id::bigint = COALESCE(:assignee_user_id::bigint, assignee_user_id::bigint)
                AND reporter_user_id::bigint = COALESCE(:reporter_user_id::bigint, reporter_user_id::bigint)
                AND CASE WHEN ARRAY[:watcher_users]::bigint[] IS NOT NULL THEN ARRAY[:watcher_users]::bigint[] <@ ARRAY(
                    SELECT w::bigint as watchers_users_info
                    FROM jsonb_object_keys(issue.watcher_users) w
                ) ELSE TRUE END
                AND (reporter_user_id = :user_id
                    OR assignee_user_id = :user_id
                    OR :user_id::text = ANY(ARRAY(
                        SELECT w::text as watchers_users_info
                        FROM jsonb_object_keys(issue.watcher_users) w
                    )))
        """
        end_of_query = """
            ORDER BY creation_time DESC
            LIMIT 1000"""
        filter_query, params = get_bind_params(form_data=form_data, result=result)
        query += (filter_query.strip() + modification_time_query + end_of_query)
        issues = self.__metadb.all(
            query,
            {
                "creation_time": creation_time,
                "issue_type_id": issue_type_id,
                "issue_status_id": issue_status_id,
                "assignee_user_id": assignee_user_id,
                "reporter_user_id": reporter_user_id,
                "watcher_users": watcher_users or [],
                "modification_time_end": modification_time_end,
                "modification_time_start": modification_time_start,
                "user_id": self.auth_user_info["auth_user_id"] if self.auth_user_info else None,
            } | params,
        )
        return [IssueGetItem(**issue) for issue in issues]

    @common_auth
    def update_issue(self, issue_id: int, issue: IssueUpdateItem) -> IssueGetItem:
        query = """
            UPDATE meta.issue SET 
                name= CASE WHEN :name::text IS NULL THEN name ELSE :name::text END,
                description= CASE WHEN :description::text IS NULL THEN description ELSE :description::text END,
                issue_priority_id= CASE WHEN :issue_priority_id::text IS NULL THEN issue_priority_id ELSE :issue_priority_id::text END,
                deadline_time= CASE WHEN :deadline_time::timestamp IS NULL THEN deadline_time ELSE :deadline_time::timestamp END,
                tags= CASE WHEN :tags::text[] IS NULL THEN tags ELSE :tags::text[] END,
                last_user_id=:user_id::bigint
            WHERE id=:issue_id::bigint
            AND (name <> :name::text
                OR description <> :description::text
                OR issue_priority_id <> :issue_priority_id::text
                OR deadline_time <> :deadline_time::timestamp
                OR tags <> :tags::text[])
        """
        self.__metadb.update(
            query,
            {
                "issue_id": issue_id,
                "name": issue.name,
                "description": issue.description,
                "issue_priority_id": issue.issue_priority_id,
                "deadline_time": issue.deadline_time,
                "tags": issue.tags,
                "user_id": self.auth_user_info["auth_user_id"] if self.auth_user_info else None,
            },
        )
        return self.get_issue(issue_id)



"""

ВНИМАНИЕ! Только для использования внутри кода SDK
Запрещено добавление общих утилит в этот файл

"""
import platform
from os.path import expanduser

import os

import json


def read_developer_settings():
    """
    Читает конфигурации разработчика с локальной машины или из переменных окружения
    При этом переменная окружения приоритетнее

    :return: dict|None
    """
    ret = read_cfg("/.rwmeta/developer_settings.json")

    env_developer_settings = os.environ.get('META_SERVICE_ACCOUNT_SECRET', None)
    if not env_developer_settings:
        env_developer_settings = os.environ.get('X-META-Developer-Settings', None)
    if env_developer_settings:
        ret = json.loads(env_developer_settings)

    return ret


def read_cfg(path) -> dict:
    """
    :param path:  example: "/.rwmeta/developer_settings.json"
    :return: dict
    """
    ret = None
    full_path = __build_path(path)
    if os.path.isfile(full_path):
        with open(full_path, 'r') as myfile:
            ret = json.loads(myfile.read())
    return ret


def write_cfg(path, value) -> None:
    """
    :param path:  example: "/.rwmeta/developer_settings.json"
    :param value: dict
    """
    full_path = __build_path(path)
    with open(full_path, 'w') as myfile:
        myfile.write(json.dumps(value))


def __build_path(path):
    if OS_NAME == "windows":
        path = path.replace("/", "\\")
    full_path = expanduser("~") + path
    return full_path


def __get_os():
    if os.name == "nt":
        return "windows"

    if platform.system() == "Darwin":
        return "macos"

    return "linux"


OS_NAME = __get_os()



import urllib.parse
from typing import Any, Optional, TypedDict, TYPE_CHECKING
from warnings import warn

import requests

from metasdk import UnexpectedError
from metasdk.exceptions import ForbiddenError, RequestError


if TYPE_CHECKING:
    from metasdk import MetaApp


class AuthUserInfo(TypedDict):
    auth_user_id: int
    company_id: int
    is_admin: bool


class AuthService:
    def __init__(self, app: "MetaApp") -> None:
        """
        Класс проверяет авторизационный токен и определяет пользователя.

        :param app: metasdk.MetaApp
        :return: None.
        """
        self.__app = app
        self.__admin_role_id: int = 6

    def verify_access_token(
        self,
        token: str,
        required_scopes: Optional[list[str]] = None,
        include_auth_info: bool = False,
    ) -> dict[str, Any]:
        """
        Проверяет токен, определяет юзера и его права.

        Проверяет корректность токена. На основе ответа от сервера верификации
        добавляет дополнительную информацию о пользователе ,совершившим запрос.

        Список доступных скоупов настраивается на странице приложения в https://console.cloud.garpun.com/
        :param token: Access Token.
        :param required_scopes: список скоупов, которые требуются сервису.
        :param include_auth_info: добавить auth_info в результат (Deprecated-параметр, будет удален).
        :return: ответ от сервера верификации и информация об авторе сервисного аккаунта.
        """
        authentication_info = self._authenticate_user(token)
        if required_scopes:
            self._check_scopes(token_scopes=authentication_info["scopes"], required_scopes=required_scopes)
        author_info = self.__set_permissions(auth_info=authentication_info)
        if include_auth_info:
            deprecation_msg = """
            include_auth_info будет удален в будущих версиях. 
            Данные от верификационного сервера теперь всегда возвращаются в ответе AuthService().verify_access_token().
            """
            warn(deprecation_msg, DeprecationWarning)
            return {**authentication_info, **author_info}
        return {"verification_server_response": authentication_info, "author_info": author_info}

    def _authenticate_user(self, token: str) -> dict[str, Any]:
        """
        Совершает запрос к серверу верификации на проверку Access токена.

        :param token: Access Token
        :return: ответ от сервера верификации.
        Возвращает скоупы и ID сервисного аккаунта.
        """
        verify_url = "oauth2/verifyJWTAccessToken"
        url = urllib.parse.urljoin(self.__app.auth_service_host, verify_url)
        response = requests.get(url, params={"assertion": token})
        if response.status_code == 200:
            return response.json()
        elif response.status_code == 403:
            raise ForbiddenError("Invalid authorization token")
        else:
            raise UnexpectedError("Unexpected error with token validation")

    def _identify_user(self, token_user_id: int) -> int:
        """
        Проверяет user_id пользователя отправившего токен.

        Если получен токен сервисного аккаунта, то возвращает user_id владельца.
        В ином случае user_id из токена
        :param meta: мета.
        :param token_user_id: user_id, полученный от авторизационного сервера.
        :return: user ID.
        """
        query = """
        SELECT author_user_id  
        FROM meta.service_account_key 
        WHERE service_account_id = :service_account_id
        """
        data = self.__app.db("meta").one(query, {"service_account_id": int(token_user_id)})
        return data["author_user_id"] if data else token_user_id

    def _check_scopes(self, token_scopes: list[str], required_scopes: list[str]) -> bool:
        if not required_scopes or not isinstance(required_scopes, list):
            raise TypeError("Invalid format of the required_scopes")

        if not token_scopes or not isinstance(token_scopes, list):
            raise TypeError("Invalid format of the token_scopes")

        if not all(scope in token_scopes for scope in required_scopes):
            raise ForbiddenError("Not enough scopes")

        return True

    def _get_user_info(self, user_id: int) -> dict[str, Any]:
        """
        Получает информацию о пльзователе по ID.

        :param meta: мета.
        :param user_id: ID пользователя в Мете.
        :return: информация о пользователе
        """
        query = "SELECT company_id, roles FROM public.users WHERE id = :user_id"
        result = self.__app.db("meta").one(query, {"user_id": int(user_id)})
        if not result:
            raise RequestError("Пользователь не найден")
        return result

    def __set_permissions(self, auth_info: dict[str, Any]) -> AuthUserInfo:
        """
        Дополняет информацию о владельце токена.

        :param auth_info: ответ от сервера верификации.
        :return: None.
        """
        auth_user_id = self._identify_user(token_user_id=int(auth_info["userId"]))
        user_info = self._get_user_info(auth_user_id)
        company_id = user_info["company_id"]
        is_admin = True if self.__admin_role_id in user_info["roles"] else False
        return {
            "auth_user_id": int(auth_user_id),
            "company_id": int(company_id),
            "is_admin": is_admin,
        }



import functools

# todo: move to AuthService!


def common_auth(func):
    """
    Декоратор используется в классах IssueService для проверки общих прав доступа
    """
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        """
        Помещает в IssueService.auth_user_info следующую информацию о пользователе:
        - auth_user_id: ID пользователя, будь то обычный пользователь или сервисный аккаунт;
        - parent_user_id: ID пользователя, если авторизуется сервисный аккаунт - ID его владельца;
        - scopes: список доступных скоупов;
        - company_id: ID компании;
        - is_admin: признак администратора.
        """
        auth_info = self.auth_service.verify_access_token(self.token, self.required_scopes)
        self.auth_user_info = {
            "auth_user_id": auth_info["verification_server_response"]["userId"],
            "parent_user_id": auth_info["author_info"]["auth_user_id"],
            "scopes": auth_info["verification_server_response"]["scopes"],
            "company_id": auth_info["author_info"]["company_id"],
            "is_admin": auth_info["author_info"]["is_admin"],
        }
        return func(self, *args, **kwargs)
    return wrapper


import json
from requests import Response


class DataResult:
    """
    Класс данных, полученных из Meta API
    Позволяет потоково получать данные по мере их поступления из api
    Служит для обработки больших массивов данных
    """

    def __init__(self, resp: Response):
        self.__resp = resp

        chunk_size = 1024 * 1024  # 1Mb

        # Если не указать chunk_size - падает после переезда в яндекс-облако с Incomplete byte read
        # http://crm.realweb.ru/browse/METAFEEDS-1281
        self.__rows_iter = self.__resp.iter_lines(chunk_size)

        self.columns = self.__line_unserialize(next(self.__rows_iter))

    def __iter__(self):
        for line in self.__rows_iter:
            yield self.__line_unserialize(line)
        self.__resp.close()

    def __line_unserialize(self, line_str):
        # IMPORTANT! orjson НЕ юзать, вызывает что-то тпа SegFault на некоторых данных
        return json.loads(line_str)


class DbService:
    def __init__(self, app):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__options = {}

    def upload(self, file_descriptor, settings):
        """
        Загружает файл в облако
        :param file_descriptor: открытый дескриптор
        :param settings: настройки загрузки
        :rtype: requests.Response
        """
        multipart_form_data = {
            'file': file_descriptor
        }
        params = {"settings": json.dumps(settings)}
        dr = self.__app.native_api_call('media', 'upload', params, self.__options, True, multipart_form_data, False, http_path="/api/meta/v1/", http_method='POST',
                                        connect_timeout_sec=60 * 10)
        return json.loads(dr.text)

    def persist_query(self, configuration):
        params = {}
        params.update(configuration)
        dr = self.__app.native_api_call('db', 'persist-query', params, self.__options)
        return json.loads(dr.text)

    def stream_query(self, configuration):
        """
        Не сохраняет ничего и ни куда.
        Мета передает данные в output,
        а python client должен их быстро вычитывать и закрывать коннект.
        Также позволяет получать список колонок
        :param configuration: dict
        :return: DataResult
        """
        params = {}
        params.update(configuration)
        dr = self.__app.native_api_call('db', 'stream-query', params, self.__options, stream=True,
                                        request_timeout_sec=60 * 60 * 3)
        return DataResult(dr)



from typing import List, Callable

from metasdk.__state import get_current_app


class Listen(object):
    """
    Декоратор для оборачивания запускаемых методов.
    С него же снимается инфа о том, какие подписки существуют.
    """

    def __init__(self, entity_id: str, code: List[str], form_pattern: List[str] = None, dispatcher_pattern: List[str] = None, receive_value: bool = False):
        """
        Регистрация метода в шине
        :param entity_id: ID сущности META
        :param code: Коды ObjectLogService META - ADD, SET, REMOVE
        :param form_pattern: Формы, на которые долэен реагировать хендлер, может быть пустым. Если пустой - значит form в event-е может быть заполненным или незаполненным,
                                а если * - то форма долэна быть заполнена, но заполнена любым значением
        :param dispatcher_pattern: Генераторы событий - используются для шины событий как клиент, который продуцирует полезну работу:
                                 запускает фоновые процессы, записыват логи, генерирует события в шину
                                 Примеры:
                                 meta.{META_APP_ALIAS} - мета-приложение
                                 appscript.{SERVICE_ID} - фоновый скрипт (включает в сещя обработчики шины событий)
                                 apiservice.{SERVICE_ID} - http апи сервисы
        :param receive_value: нужно ли добавлять value в событие
        """
        self.entity_id = entity_id
        self.code = code
        self.form_pattern = form_pattern
        self.dispatcher_pattern = dispatcher_pattern
        self.receive_value = receive_value

    def __call__(self, original_func):
        get_current_app().event_bus.add_listener(self, original_func)

        def wrappee(*args, **kwargs):
            return original_func(*args, **kwargs)

        return wrappee


class Listener(object):
    """
    Хранилище функции и настройки
    """

    def __init__(self, listen: Listen, handler_func: Callable):
        self.listen = listen
        self.handler_func = handler_func


class Event(object):
    """
    Объект события с его контентом.
    На основе данных таска запускатора
    """

    def __init__(self, task):
        self.__task = task

        edata = task['data']['event']
        self.event_id = edata.get("eventId")
        self.user_id = int(edata['userId'])
        self.entity_id = str(edata['entityId'])
        self.object_id = str(edata['objectId'])
        self.code = str(edata['code'])
        self.form = str(edata['form']) if edata.get('form') else None
        self.dispatcher = str(edata['dispatcher']) if edata.get('dispatcher') else None
        self.value = edata.get('value')


class EventBus(object):
    """
    Сама Шина. Она же хранит всю карту слушателей
    """
    listener = Listen

    def __init__(self, app) -> None:
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__listeners_map = {}

    def add_listener(self, listen, handler_func):
        entity_id = listen.entity_id
        self.__listeners_map.setdefault(entity_id, [])
        listener = Listener(listen, handler_func)
        self.__listeners_map[entity_id].append(listener)

    def get_listeners(self):
        for listeners in self.__listeners_map.values():
            for listener in listeners:
                yield listener

    def accept(self, task):
        event = Event(task)
        # listener: type Listener
        for listener in self.__listeners_map.get(event.entity_id, []):
            if event.code not in listener.listen.code:
                continue

            if not self.__contains_in_patterns(event.form, listener.listen.form_pattern):
                continue

            if not self.__contains_in_patterns(event.dispatcher, listener.listen.dispatcher_pattern):
                continue

            try:
                self.__app.log.set_entity("event", {
                    "dispatcher": event.dispatcher,
                    "eventId": event.event_id,
                    "userId": event.user_id,
                    "entityId": event.entity_id,
                    "objectId": event.object_id,
                    "code": event.code,
                    "form": event.form,
                })
                listener.handler_func(event)
            finally:
                self.__app.log.set_entity("event", None)

    def __contains_in_patterns(self, in_str, pattern_list):
        if not pattern_list:
            return True

        if not in_str:
            # если нет формы, то паттерны обязаны содержать хотябы один, целиком состоящий из *
            # это будет означать, что любые form годятся для хендлера
            return '*' in pattern_list

        for pattern in pattern_list:
            if pattern.startswith("*"):
                # есть подобие регулярки. Например qwe*
                if in_str.endswith(pattern[1:]):
                    return True
            elif pattern.endswith("*"):
                # есть подобие регулярки
                if in_str.startswith(pattern[:-1]):
                    return True
            else:
                # регулярки нет, можно проверять точным вхождением
                if in_str == pattern:
                    return True
        return False



import json


class ObjectLogService:
    """
    Служба журналирования логов по объектам
    """

    def __init__(self, app):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__options = {}

    def log(self, record):
        """
        Делает запись по объекту в журнале
        """
        record['dispatcher'] = self.__app.dispatcher_name
        if 'value' in record and record['value']:
            if not isinstance(record['value'], dict):
                raise ValueError("ObjectLogService expected dict in log record value field")
            record['jsonValue'] = json.dumps(record['value'])
            record.pop('value')
        body_value = {
            "record": record
        }
        return self.__app.native_api_call('object-log', 'log', body_value, self.__options, False, None, False, http_path="/api/meta/v1/", http_method='POST')



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Optional, ParamSpec
import json


class Message(ABC):
    """Абстрактный класс сообщения"""
    messenger_id = "abstract"

    @abstractmethod
    def generate_unique_id(self) -> str:
        """
        Генерирует уникальный идентификатор сообщения по его атрибутам.

        Абстрактный метод. В наследниках необходимо переопределить.
        :return: Уникальный идентификатор сообщения.
        """
        pass

    def __init__(
        self,
        user_id: str,
        body: str,
        template: str = "",
        unique_id: Optional[str] = None,
        author_user_id: Optional[int] = None,
        sender_id: Optional[int] = None,
        **kwargs: ParamSpec.kwargs,
    ) -> None:
        """
        Инициализирует экземпляр класса.

        :param user_id: Идентификатор пользователя-адресата.
        :param body: Строка с телом отправляемого сообщения.
        :param template: Идентификатор шаблона сообщения.
        :param unique_id: Уникальный идентификатор сообщения.
        :param kwargs: Опциональные параметры, актуальные для отдельных каналов отправки сообщений.
        :param sender: Идентификатор отправителя.
        :return: None.
        """
        self._messenger = self.messenger_id
        self._user_id = user_id
        self._body = body
        self._template = template
        self._custom_unique_id = unique_id
        self._author_user_id = author_user_id
        self._sender_id = sender_id
        self._kwargs = kwargs

    def to_dict(self) -> dict[str, int | str]:
        """
        Возвращает словарь с данными сообщения.

        :return: Словарь с данными сообщения. Удобно передавать методам DbQueryService.
        """

        return {
            "messenger": self._messenger,
            "user_id": self._user_id,
            "body": self._body,
            "template": self._template,
            "unique_id": self._custom_unique_id or self.generate_unique_id(),
            "author_user_id": self._author_user_id,
            "sender_info": json.dumps({"id": self._sender_id}),
        }


class WebhookMessage(Message):
    """Класс сообщения, отправляемого вебхуком"""
    messenger_id = "webhook"

    def generate_unique_id(self) -> str:
        """
        Генерирует уникальный идентификатор сообщения по его атрибутам.

        В kwargs экземпляра можно передать параметр "uniqualizer",
        и он будет добавлен в начало unique_id. Так, скажем, можно проверять,
        был ли уже отправлен вебхук по конкретной записи из meta.object_log. 
        :return: Уникальный идентификатор сообщения.
        """
        unique_id_head = "webhook"
        if uniqualizer := self._kwargs.get("uniqualizer"):
            unique_id_head += f"_{uniqualizer}"       
        timestamp = datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
        return f"{unique_id_head}_{self._user_id}_{timestamp}"


class TelegramMessage(Message):
    """Класс сообщения, отправляемого через Telegram"""
    messenger_id = "telegram"

    def __init__(
        self, 
        user_id: str, 
        body: str, 
        sender_id: str = "80a5b860-683e-4afd-9733-572826bbac1e",  # feedsgarpunbot
        template: str = "", 
        unique_id: Optional[str] = None, 
        author_user_id: Optional[int] = None, 
        **kwargs: ParamSpec.kwargs,
    ):
        """
        :param sender_id: Идентификатор подключения из meta.ex_access к телеграм боту
        :return: None.
        """
        super().__init__(user_id, body, template, unique_id, author_user_id, sender_id, **kwargs)

    def generate_unique_id(self) -> str:
        """
        Генерирует уникальный идентификатор сообщения по его атрибутам.

        :return: Уникальный идентификатор сообщения.
        """
        timestamp = datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
        return f"custom_message_{self._user_id}_{timestamp}_{self._sender_id}"


class MailService:
    def __init__(self, app):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__options = {}
        self.__data_get_cache = {}
        self.__metadb = app.db("meta")
        self.log = app.log

    def submit_mail(
        self,
        send_from: str,
        send_to: str,
        subject: str,
        body: str,
        unique_id: Optional[str] = None,
    ) -> None:
        """
        Добавляет письмо в очередь на отправку (фактически — запись в таблицу meta.mail).

        :param send_from: Отправитель.
        :param send_to: Получатель.
        :param subject: Тема письма.
        :param body: Тело письма. Можно с HTML.
        :param unique_id: Уникальный идентификатор письма.
            Лучше всего подойдет md5 + человекочитаемый префикс.
            Письмо с существующим unique_id не будут добавлено.
        :return: None.
        """
        self.__metadb.update(
            """
                INSERT INTO meta.mail(
                    "template",
                    "from",
                    "to",
                    "subject",
                    "body",
                    "attachments",
                    "unique_id"
                )
                VALUES ('meta', :send_from, :send_to, :subject, :body, null, :unique_id)
                ON CONFLICT (unique_id) DO NOTHING
            """,
            {
                "send_from": send_from,
                "send_to": send_to,
                "subject": subject,
                "body": body,
                "unique_id": unique_id,
            },
        )

    def submit_message(self, message: Message) -> None:
        """
        Добавляет сообщение в очередь на отправку (фактически — запись в таблицу meta.messenger).

        :param message: Экземпляр класса-наследника Message.
        :return: None.
        """
        self.__metadb.update(
            """
                INSERT INTO meta.messenger(
                    messenger,
                    user_id,
                    body,
                    unique_id,
                    author_user_id, 
                    sender_info
                )
                VALUES (
                    :messenger::text,
                    :user_id::text,
                    :body::text,
                    :unique_id::text,
                    COALESCE(:author_user_id, valera_user_id())::bigint,
                    :sender_info::jsonb
                )
                ON CONFLICT (unique_id) DO NOTHING
            """,
            message.to_dict(),
        )
